% Stelle DGL-System auf fuer die Waermeleitungsgleichung
% (s. VL Kap. 5, Bsp. (14))
M = 25; % 25, 50, 100 Gitter Groesse
x = linspace(0, 1., M + 1);
dx = 1./M;
grid = numgrid('S', M + 1);
A = - delsq(grid)/dx^2;
u = 25.*ones(M + 1, M + 1);
u(:,1) = 0.; u(:,M+1) = 0.; u(1,:) = 0.; u(M+1,:) = 0;

% Zeitintegration
tbeg = 0.;
tend = 0.1;
y0 = reshape(u(2:M,2:M), (M - 1)^2, 1);
f = @(t, y) A*y;

for sol=1:4
  tic
  if     sol == 1 % expliziter Euler
    disp('Loeser: expliziter Euler')
    k  = dx^2/12.; % groesst moeglicher Zeitschritt gemaess Eigenwerte von A
    N = ceil((tend - tbeg)/k);
    dt = (tend - tbeg)/N;
    t = tbeg;
    y = y0;
    for n=1:N
      t = t + dt;
      y = y + dt*f(t,y);
    end

  elseif sol == 2 % MATLAB's ode45
    disp('Loeser: MATLAB ode45')
    tspan = [tbeg, 0.5*(tbeg + tend), tend];
    opts = odeset('Stats', 'on');
    [t, yy] = ode45(f, tspan, y0, opts);
    t = t(end);   % wir behalten nur den
    y = yy(3, :); % letzten Zeitschritt

  elseif sol == 3 % MATLAB's ode23s
    disp('Loeser: MATLAB ode23s')
    tspan = [tbeg, 0.5*(tbeg + tend), tend];
    opts = odeset('Jacobian', A, 'Stats', 'on');
    % opts = odeset('Stats', 'on'); % VORSICHT: DIES KANN LANGE GEHEN!
    [t, yy] = ode23s(f, tspan, y0, opts);
    t = t(end);   % wir behalten nur den
    y = yy(3, :); % letzten Zeitschritt

  elseif sol == 4 % Crank-Nicolson (aka implizite Trapezregel)
    disp('Loeser: Crank-Nicolson (aka implizite Trapezregel)')
    k = dx/12.; % Obwohl dieser Loeser A-Stabilil ist, kann es zu stoerenden
                % (aber nicht wachsenden) Oszillationen kommen fuer allzu grosse
                % Zeitschritte
    N = ceil((tend - tbeg)/k);
    dt = (tend - tbeg)/N;
    t = tbeg;
    y = y0;
    ImhA = speye((M - 1)^2) - 0.5*dt*A; % I - dt*A
    IphA = speye((M - 1)^2) + 0.5*dt*A; % I + dt*A
    for n=1:N
      t = t + dt;
      % y = ImhA \ (IphA*y); % verwendet einen sog. direkten Loeser welcher das
      %                      % lineare Gleichungssystem (LGS) mit varianten der
      %                      % Gauss-Elimination loest
      [y, flag] = pcg(ImhA, IphA*y); % noch schneller geht's mit einem sog.
                                     % iterativen Loeser welcher das LGS
                                     % naeherungsweise loest
    end

  end
  toc
end


% Plotte Loesung
u(2:M,2:M) = reshape(y, M - 1, M - 1);
surf(x, x, u);
colorbar();
xlabel('x');
ylabel('y');
zlabel('u');
title(sprintf('Temperature zur Zeit t = %.2f', t));
